import logging

import attr
from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.remote.webdriver import WebDriver

from .config import Browser

logger = logging.getLogger('scalpel')


@attr.s
class SeleniumDriverMixin:
    _driver: WebDriver = attr.ib(init=False, repr=False)

    @_driver.default
    def _get_driver(self) -> WebDriver:
        if self.config.selenium_browser is Browser.FIREFOX:
            logger.debug('returning firefox driver')
            options = FirefoxOptions()
            options.headless = True
            driver = webdriver.Firefox(
                options=options, executable_path=self.config.selenium_driver_executable_path,
                service_log_path=self.config.selenium_driver_log_file
            )
        else:
            logger.debug('returning chrome driver')
            options = ChromeOptions()
            options.add_argument('--headless')
            driver = webdriver.Chrome(
                options=options, executable_path=self.config.selenium_driver_executable_path,
                service_log_path=self.config.selenium_driver_log_file
            )
        driver.implicitly_wait(self.config.selenium_find_timeout)
        return driver



import logging
from typing import Dict, Optional

import attr
import httpx
import parsel
from rfc3986 import uri_reference
from selenium.webdriver.remote.webdriver import WebDriver

logger = logging.getLogger('scalpel')


@attr.s(kw_only=True)
class BaseStaticResponse:
    _url: str = attr.ib(default='', validator=attr.validators.optional(attr.validators.instance_of(str)))
    _text: str = attr.ib(default='', validator=attr.validators.instance_of(str))
    _httpx_response: Optional[httpx.Response] = attr.ib(
        default=None, validator=attr.validators.optional(attr.validators.instance_of(httpx.Response))
    )
    _selector: parsel.Selector = attr.ib(init=False)

    @property
    def url(self) -> str:
        """The url associated to the response object."""
        if self._url:
            _url = self._url
        else:
            _url = f'{self._httpx_response.url}'
        logger.debug('returning response url: %s', _url)
        return _url

    @property
    def headers(self) -> Dict[str, str]:
        """A `dict` of http headers in case of an HTTP url. Empty `dict` otherwise."""
        if self._url:
            _headers = {}
        else:
            _headers = dict(self._httpx_response.headers)
        logger.debug('returning response headers:\n %s', _headers)
        return _headers

    @property
    def text(self) -> str:
        """The string content associated to the url."""
        if self._text:
            _text = self._text
        else:
            _text = self._httpx_response.text
        logger.debug('returning response text content:\n %s', _text)
        return _text

    @property
    def content(self) -> bytes:
        """The bytes content associated to the url."""
        if self._text:
            _content = self._text.encode(errors='replace')
        else:
            _content = self._httpx_response.content
        logger.debug('returning response byte content:\n %s', _content)
        return _content

    @property
    def cookies(self) -> Dict[str, str]:
        """A `dict` of cookies associated to the response in case of an HTTP url. Empty `dict` otherwise."""
        if self._url:
            _cookies = {}
        else:
            _cookies = dict(self._httpx_response.cookies)
        logger.debug('returning response cookies: %s', _cookies)
        return _cookies

    @_selector.default
    def _get_selector(self) -> parsel.Selector:
        logger.debug('creating parsel selector with text:\n %s', self.text)
        return parsel.Selector(self.text)

    def css(self, query: str) -> parsel.SelectorList:
        """
        Applies CSS rules to select DOM elements.

        **Parameters:**

        * **query:** The CSS rule used to select DOM elements.

        **Returns:** `parsel.SelectorList`
        """
        logger.debug('selecting content using css selector: %s', query)
        return self._selector.css(query)

    def xpath(self, query: str) -> parsel.SelectorList:
        """
        Applies XPath rules to select DOM elements.

        **Parameters:**

        * **query:** The XPath rule used to select DOM elements.

        **Returns:** `parsel.SelectorList`
        """
        logger.debug('selecting content using xpath selector: %s', query)
        return self._selector.xpath(query)

    def _get_absolute_url(self, url: str) -> str:
        """
        This method returns absolute url from local or http urls.
        """
        current_uri = uri_reference(self._url) if self._url else uri_reference(str(self._httpx_response.url))
        given_uri = uri_reference(url)
        if given_uri.is_absolute():
            _url = url
        else:
            uri = given_uri.resolve_with(current_uri)
            uri = uri.copy_with(fragment=None)
            _url = uri.unsplit()
        logger.debug('returning computed absolute url: %s', _url)
        return _url


@attr.s(kw_only=True)
class BaseSeleniumResponse:
    driver: WebDriver = attr.ib(validator=attr.validators.instance_of(WebDriver))
    handle: str = attr.ib(validator=attr.validators.instance_of(str))

    def _get_absolute_url(self, url: str) -> str:
        """
        This method returns absolute url from local or http urls.
        """
        uri = uri_reference(url)
        if uri.is_absolute():
            _url = url
        else:
            # this is probably useless because I checked with firefox and chrome and both
            # computes absolute urls when they encounter a relative one, but to remove any doubt, let's do this
            current_uri = uri_reference(self.driver.current_url)
            uri = uri.resolve_with(current_uri)
            uri = uri.copy_with(fragment=None)
            _url = uri.unsplit()
        logger.debug('returning computed absolute url: %s', _url)
        return _url



import logging
from pathlib import Path
from typing import Dict, Union, List
from urllib.robotparser import RobotFileParser

import anyio
import attr
import httpx

from scalpel.core.robots import RobotsMixin

logger = logging.getLogger('scalpel')


@attr.s
class RobotsAnalyzer(RobotsMixin):
    # order is important, _http_client can use the _user_agent info to instantiate itself
    _user_agent: str = attr.ib()
    _robots_cache: Path = attr.ib()
    _robots_mapping: Dict[str, Path] = attr.ib(factory=dict)
    _http_client: httpx.AsyncClient = attr.ib()
    _robots_parser: RobotFileParser = attr.ib(init=False, factory=RobotFileParser)
    _delay_mapping: Dict[str, Union[int, float]] = attr.ib(init=False, factory=dict)

    @_http_client.default
    def _get_default_client(self) -> httpx.AsyncClient:
        logger.debug('returning default http client with user agent: %s', self._user_agent)
        headers = {'User-Agent': self._user_agent}
        return httpx.AsyncClient(headers=headers)

    @staticmethod
    async def _create_robots_file(robots_path: Path, content: str) -> None:
        logger.debug('creating robots file at path: %s', robots_path)
        async with await anyio.open_file(robots_path, 'w') as f:
            await f.write(content)

    @staticmethod
    async def _get_robots_lines(path: Path) -> List[str]:
        logger.debug('getting robots file lines at path: %s', path)
        async with await anyio.open_file(path) as f:
            return await f.readlines()

    async def can_fetch(self, url: str) -> bool:
        httpx_url = httpx.URL(url)
        host = httpx_url.host
        robots_url = httpx_url.copy_with(path='/robots.txt')

        if host not in self._robots_mapping:
            try:
                response = await self._http_client.get(f'{robots_url}')
            except httpx.ConnectTimeout:
                logger.info('cannot connect to host % to get robots.txt file, returning False', robots_url.host)
                return False
            if response.status_code in (401, 403):
                logger.info(
                    'access to %s is %s, returning False', f'{robots_url}',
                    httpx.codes.get_reason_phrase(response.status_code)
                )
                return False
            elif httpx.codes.is_error(response.status_code):
                logger.info(
                    'trying to access %s returns a %s error status code, returning True', f'{robots_url}',
                    response.status_code
                )
                return True
            else:
                robot_path = self._robots_cache / host
                await self._create_robots_file(robot_path, response.text)
                self._robots_mapping[host] = robot_path.absolute()

        robot_lines = await self._get_robots_lines(self._robots_mapping[host])
        self._robots_parser.parse(robot_lines)
        is_fetchable = self._robots_parser.can_fetch(self._user_agent, url)
        logger.info('after analyzing %s file, returning value is %s', f'{robots_url}', is_fetchable)
        return is_fetchable

    async def get_request_delay(self, url: str, delay: Union[int, float]) -> Union[int, float]:
        host = httpx.URL(url).host

        if host in self._delay_mapping:
            logger.debug('returning caching value %s', self._delay_mapping[host])
            return self._delay_mapping[host]

        if host not in self._robots_mapping:
            is_fetchable = await self.can_fetch(url)
            if not is_fetchable:
                self._delay_mapping[host] = -1
                logger.debug('url %s is not fetchable, returning negative value', url)
                return -1

        return self._get_request_delay(host, url, self._robots_parser, self._delay_mapping, delay)

    async def close(self) -> None:
        await self._http_client.aclose()



"""Utilities to read / write files and manage different IO buffers"""
from io import BufferedReader, open
from typing import IO, AnyStr, Union, List, Optional, Iterator, Any, TypeVar, Callable

import attr
from gevent import get_hub
from gevent.threadpool import ThreadPool

File = TypeVar('File')


@attr.s
class AsyncFile:
    """
    A wrapper around builtins io objects like `io.StringIO` or `io.BufferedReader` running blocking operations like
    `read` or `write` in a threadpool to make it gevent cooperative.
    """
    _wrapper: Union[IO, BufferedReader] = attr.ib()
    _pool: ThreadPool = attr.ib(init=False)

    @_pool.default
    def _get_pool(self) -> ThreadPool:
        return get_hub().threadpool

    def read(self, size: int = -1) -> AnyStr:
        return self._pool.spawn(self._wrapper.read, size).get()

    def read1(self, size: int = -1) -> AnyStr:
        return self._pool.spawn(self._wrapper.read1, size).get()

    def readline(self, size: int = -1) -> AnyStr:
        return self._pool.spawn(self._wrapper.readline, size).get()

    def readlines(self, hint: int = -1) -> List[AnyStr]:
        return self._pool.spawn(self._wrapper.readlines, hint).get()

    def readinto(self, b: Union[bytes, bytearray, memoryview]) -> int:
        return self._pool.spawn(self._wrapper.readinto, b).get()

    def readinto1(self, b: Union[bytes, bytearray, memoryview]) -> int:
        return self._pool.spawn(self._wrapper.readinto1, b).get()

    def seek(self, offset: int, whence: int = 0) -> int:
        return self._pool.spawn(self._wrapper.seek, offset, whence).get()

    def tell(self) -> int:
        return self._pool.spawn(self._wrapper.tell).get()

    def write(self, s: AnyStr) -> int:
        return self._pool.spawn(self._wrapper.write, s).get()

    def writelines(self, lines: List[AnyStr]) -> None:
        return self._pool.spawn(self._wrapper.writelines, lines).get()

    def truncate(self, size: Optional[int] = None) -> int:
        return self._pool.spawn(self._wrapper.truncate, size).get()

    def peek(self, size: Optional[int] = None) -> bytes:
        return self._pool.spawn(self._wrapper.peek, size).get()

    def flush(self) -> None:
        return self._pool.spawn(self._wrapper.flush).get()

    def close(self) -> None:
        return self._pool.spawn(self._wrapper.close).get()

    def __enter__(self) -> 'AsyncFile':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        return self.close()

    def __iter__(self) -> Iterator[AnyStr]:
        while True:
            line = self.readline()
            if not line:
                break
            yield line

    def __getattr__(self, item: str) -> Any:
        return getattr(self._wrapper, item)


def _has(file: Union[IO, BufferedReader], attribute: str) -> bool:
    return hasattr(file, attribute) and callable(getattr(file, attribute))


def wrap_file(file: Union[IO, BufferedReader]) -> AsyncFile:
    """
    This function wraps any file object in a wrapper that provides an asynchronous (or gevent cooperative)
    file object interface.

    **Parameters:**

    * **file:** A file-like object.

    **Returns:** An `AsyncFile` object.

    Usage:

    ```
    from io import StringIO
    from scalpel.green import wrap_file

    s = StringIO()
    async_s = wrap_file(s)
    assert 5 == async_s.write('hello')
    assert 'hello' == async_s.getvalue()
    ```
    """
    if not _has(file, 'close'):
        raise TypeError(f'{file} does not implement close method which is mandatory for a file-like object')

    if not _has(file, 'read') or not _has(file, 'write'):
        raise TypeError(f'{file} does not implement read or write method which is mandatory for a file-like object')

    return AsyncFile(file)


def open_file(
        file: File,
        mode: str = 'r',
        buffering: int = -1,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        closefd: bool = True,
        opener: Callable = None
) -> AsyncFile:
    """
    An asynchronous version of the builtin `open` function running blocking operation in a threadpool.

    **Parameters:**
    The parameters are exactly the same as those passed to the builtin `open` function. You can check the official
    documentation to understand their meaning.

    **Returns:** An `AsyncFile` object.

    Usage:

    ```
    from scalpel.green import open_file

    with open_file('hello.txt', 'w') as f:
        f.write('hello world')

    with open_file('hello.txt') as f:
        print(f.read())  # 'hello world'
    ```
    """
    threadpool = get_hub().threadpool
    file = threadpool.spawn(open, file, mode, buffering, encoding, errors, newline, closefd, opener).get()
    return wrap_file(file)



import logging
import random
import re
import tempfile
import uuid
from enum import Enum, auto
from importlib import import_module
from pathlib import Path
from typing import List, Callable, Any, Dict, Union, Optional

import attr
from configuror import Config
from fake_useragent import UserAgent, FakeUserAgentError

from .message_pack import datetime_encoder, datetime_decoder

logger = logging.getLogger('scalpel')


def check_value_greater_or_equal_than_0(_, attribute: attr.Attribute, value: int) -> None:
    if value < 0:
        message = f'{attribute.name} must be a positive integer'
        logger.exception(message)
        raise ValueError(message)


def check_max_delay_greater_or_equal_than_min_delay(instance: 'Configuration', attribute: attr.Attribute,
                                                    value: int) -> None:
    if instance.min_request_delay > value:
        message = f'{attribute.name} must be greater or equal than min_request_delay'
        logger.exception(message)
        raise ValueError(message)


def check_file_presence(_, attribute: attr.Attribute, filename: str) -> None:
    path = Path(filename)
    if not path.exists():
        message = f'File {filename} does not exist'
        logger.exception(f'attribute {attribute.name} does not have a valid path: {message}')
        raise FileNotFoundError(message)


def check_driver_presence(config: 'Configuration', attribute: attr.Attribute, filename: str) -> None:
    if filename in ['chromedriver', 'geckodriver']:
        return
    check_file_presence(config, attribute, filename)


def validate_robots_folder(_, attribute: attr.Attribute, path: Path) -> None:
    if not path.exists():
        message = f'{attribute.name} does not exist'
        logger.exception(message)
        raise FileNotFoundError(message)

    dummy_file = path / 'dummy_file'
    try:
        dummy_file.write_text('hello')
    except PermissionError:
        logger.exception(f'Cannot write file in {path}')
        raise
    try:
        dummy_file.read_text()
    except PermissionError:
        logger.exception(f'Cannot read file in {path}')
        raise
    dummy_file.unlink()


def check_file_can_be_created(_, _attribute: attr.Attribute, value: str) -> None:
    if value is not None:
        p = Path(value)
        # touch helps to see if a file can be created with the given path
        p.touch()
        # we don't want to have a created file if other attributes validation failed
        p.unlink()


# I could just use return type "Any" but I want to insist on the fact that the function must
# first return a boolean and in the other cases, the value given at input
def bool_converter(value: Any) -> Union[bool, Any]:
    if not isinstance(value, str):
        logger.debug('%s is not a string, returned it as it is', value)
        return value

    if value.lower() in ['1', 'true', 'yes', 'y']:
        logger.debug('converts %s to True', value)
        return True
    elif value.lower() in ['0', 'false', 'no', 'n']:
        logger.debug('converts %s to False', value)
        return False
    else:
        message = f'{value} does not represent a boolean'
        logger.exception(message)
        raise ValueError(message)


def get_callable_from_string(callable_string: str) -> Callable:
    parts = callable_string.split('.')
    module_name = '.'.join(parts[:-1])
    module = import_module(module_name)
    return getattr(module, parts[-1])


# The same logic as the bool converter applies to the type of return
def callable_list_converter(value: Any) -> Union[List[Callable], Any]:
    if isinstance(value, list):
        if not all(isinstance(item, str) for item in value):
            logger.debug('not all items in the list are a string, returned it as it: %s', value)
            return value
        str_callable_list = value
    elif isinstance(value, str):
        str_callable_list = re.split(r',\s*|;\s*|:\s*|\s+', value)
    else:
        logger.debug('%s is not a string or a list of strings, returned it as it is', value)
        return value

    callables = []
    for str_callable in str_callable_list:
        callables.append(get_callable_from_string(str_callable))

    logger.debug('returning callables: %s', callables)
    return callables


def msgpack_converter(value: Any) -> Union[Callable, Any]:
    if not isinstance(value, str):
        logger.debug(f'{value} is not a string, returning it as it')
        return value
    return get_callable_from_string(value)


def str_converter(value: Any) -> Optional[str]:
    if value is None:
        return
    if isinstance(value, Path):
        return str(value.absolute())
    return str(value)


class Browser(Enum):
    """An enum with different browser values."""
    FIREFOX = auto()
    CHROME = auto()


def browser_converter(value: Any) -> Any:
    if isinstance(value, str):
        upper_value = value.upper()
        if upper_value == Browser.FIREFOX.name:
            return Browser.FIREFOX
        if upper_value == Browser.CHROME.name:
            return Browser.CHROME
    return value


positive_int_validators = [attr.validators.instance_of(int), check_value_greater_or_equal_than_0]
max_delay_validators = [*positive_int_validators, check_max_delay_greater_or_equal_than_min_delay]
positive_float_validators = [attr.validators.instance_of(float), check_value_greater_or_equal_than_0]
middleware_validator = attr.validators.deep_iterable(
    member_validator=attr.validators.is_callable(),
    iterable_validator=attr.validators.instance_of((list, tuple))
)
backup_filename_validators = [attr.validators.instance_of(str), check_file_can_be_created]
selenium_path_validators = [attr.validators.optional(attr.validators.instance_of(str)), check_file_can_be_created]


@attr.s(frozen=True)
class Configuration:
    """
    Configure variables for your spider.

    **Parameters:**

    * **min_request_delay:** The minimum delay to wait between two http requests. Defaults to 0s.

    * **max_request_delay:** The maximum delay to wait between two http requests. Defaults to 0s.

    * **fetch_timeout:** The timeout to fetch http resources using the inner
    [httpx](https://www.python-httpx.org/) client. Defaults to 5s.

    * **selenium_find_timeout:** The timeout for selenium driver to find an element in a page. Defaults to 10s.

    * **selenium_driver_log_file:** The file where the browser log debug messages. Defaults to *driver.log*.
    If you want to not create one, just pass `None`.

    * **selenium_browser:** The browser to use with the selenium spider. You can use the `Browser` enum to specify the
    value. Possible values are `Browser.FIREFOX` and `Browser.CHROME`. Defaults to `Browser.FIREFOX`.

    * **selenium_driver_executable_path:** The path to the browser driver. Defaults to *geckodriver* if
    `Browser.FIREFOX` is selected as *selenium_browser*, otherwise defaults to *chromedriver*.

    * **user_agent:** The user agent to fake. Mainly useful for the static spider. Defaults to a random value provided
    by [fake-useragent](https://pypi.org/project/fake-useragent/) and if it does not work, fallback to
    *Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2225.0 Safari/537.36*

    * **follow_robots_txt:** Decide whether or not the spider should follow robots.txt rules on the website you are
    scraping. Defaults to `False`.

    * **robots_cache_folder:** A folder to cache content of different website robots.txt file to avoid retrieving
    it each time you want to analyze an html page. Default to the system temporary directory.

    * **backup_filename:** The filename were scraped items will be written. If you don't want one, simple pass `None`.
    Defaults to *backup-{uuid}.mp* where uuid is a `uuid.uuid4` string value. Note that values inserted in this file
    are streamed using `msgpack`. Look at the documentation to see how to use it.

    * **response_middlewares:** A list of callables that will be called with the callable that fetch the http resource.
    This parameter is only useful for the **static spider**. Defaults to an empty list.

    * **item_processors:** A list of callables that will be called with a scraped item. Defaults to an empty list.

    * **msgpack_encoder:** A callable that will be called when `msgpack` serializes an item.
    Defaults to `scalpel.datetime_encoder`.

    * **msgpack_decoder:** A callable that will be called when `msgpack` deserializes an item.
    Defaults to `scalpel.datetime_decoder`.

    Usage:

    ```
    from scalpel import Configuration, Browser
    config = Configuration(
        min_request_delay=1, max_request_delay=3, follow_robots_txt=True, selenium_browser=Browser.CHROME
    )
    ```
    """
    min_request_delay: int = attr.ib(default=0, converter=int, validator=positive_int_validators)
    max_request_delay: int = attr.ib(default=0, converter=int, validator=max_delay_validators)
    fetch_timeout: float = attr.ib(default=5.0, converter=float, validator=positive_float_validators)
    selenium_find_timeout: float = attr.ib(default=10.0, converter=float, validator=positive_float_validators)
    selenium_driver_log_file: Optional[str] = attr.ib(
        converter=str_converter, default='driver.log', validator=selenium_path_validators
    )
    selenium_browser: Browser = attr.ib(
        default=Browser.FIREFOX, converter=browser_converter, validator=attr.validators.in_(Browser)
    )
    # default value of this attribute depends if selenium browser, so the order is important here
    selenium_driver_executable_path: str = attr.ib(
        converter=str_converter, validator=[attr.validators.instance_of(str), check_driver_presence]
    )
    user_agent: str = attr.ib(validator=attr.validators.instance_of(str))
    follow_robots_txt: bool = attr.ib(
        default=False, converter=bool_converter, validator=attr.validators.instance_of(bool)
    )
    robots_cache_folder: Path = attr.ib(converter=Path, validator=validate_robots_folder)
    backup_filename: str = attr.ib(validator=backup_filename_validators)
    response_middlewares: List[Callable] = attr.ib(
        repr=False, converter=callable_list_converter, factory=list, validator=middleware_validator
    )
    item_processors: List[Callable] = attr.ib(
        repr=False, converter=callable_list_converter, factory=list, validator=middleware_validator
    )
    msgpack_encoder: Callable = attr.ib(
        repr=False, converter=msgpack_converter, default=datetime_encoder, validator=attr.validators.is_callable()
    )
    msgpack_decoder: Callable = attr.ib(
        repr=False, converter=msgpack_converter, default=datetime_decoder, validator=attr.validators.is_callable()
    )

    @user_agent.default
    def _get_default_user_agent(self) -> str:
        try:
            ua = UserAgent()
            user_agent = ua.random
            logger.debug('returning a random user agent: %s', user_agent)
            return user_agent
        except FakeUserAgentError:
            # for the fallback, I use a recent version found on http://useragentstring.com/
            # not sure if this is the best strategy but we will stick with it for now
            fallback = 'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) ' \
                       'Chrome/41.0.2225.0 Safari/537.36'
            logger.debug('returning fallback value for user agent: %s', fallback)
            return fallback

    @robots_cache_folder.default
    def _get_robots_cache_folder(self) -> Path:
        temp_dir = Path(tempfile.mkdtemp(prefix='robots_'))
        logger.debug('returning default created temporary directory: %s', temp_dir)
        return temp_dir

    @backup_filename.default
    def _get_backup_filename(self) -> str:
        name = f'backup-{uuid.uuid4()}.mp'
        logger.debug('returning computed backup filename: %s', name)
        return name

    @selenium_driver_executable_path.default
    def _get_driver_executable_path(self) -> str:
        executable = 'geckodriver' if self.selenium_browser is Browser.FIREFOX else 'chromedriver'
        # I don't use self.selenium_browser.name attribute because some tests fail here when testing browser attribute
        # with a string
        logger.debug(
            'returning default executable path to %s since browser selected is %s', executable,
            self.selenium_browser
        )
        return executable

    @property
    def request_delay(self) -> int:
        """
        A read-only property which is a random value between `min_request_delay` and `max_request_delay`
        (both sides included) and used to wait between two http requests.
        """
        # for bandit, using random module to generate pseudo-random values is not a good
        # idea for cryptography / security purposes, but since we are not in this case, we just
        # ignore this warning.
        # More about the warning: https://bandit.readthedocs.io/en/latest/blacklists/blacklist_calls.html#b311-random
        delay = random.randint(self.min_request_delay, self.max_request_delay)  # nosec
        logger.debug('returning computed request delay: %s s', delay)
        return delay

    @staticmethod
    def _get_dict_with_lower_keys(data: Dict[str, Any]) -> Dict[str, Any]:
        data = {key.lower(): value for key, value in data.items()}
        logger.debug('returning dict with lower keys: %s', data)
        return data

    @classmethod
    def _scalpel_attributes(cls, data: Dict[str, Any]) -> Dict[str, Any]:
        data_key = 'scalpel'
        attributes = {}

        if data_key not in data:
            logger.debug('no namespace "scalpel" in %s, returning empty attributes', data)
            return attributes

        data = cls._get_dict_with_lower_keys(data[data_key])
        for attribute in attr.fields(cls):
            if attribute.name != '_config' and attribute.name in data:
                attributes[attribute.name] = data[attribute.name]
        logger.debug('returning scalpel attributes: %s', attributes)
        return attributes

    @staticmethod
    def _check_file(config_file: Union[Path, str], file_type: str) -> None:
        if not isinstance(config_file, (Path, str)):
            error_message = f'{file_type} file must be of type Path or str but you provided {type(config_file)}'
            logger.exception(error_message)
            raise TypeError(error_message)

        config_file = Path(config_file)
        if not config_file.is_file():
            error_message = f'file {config_file} does not exist'
            logger.exception(error_message)
            raise FileNotFoundError(error_message)

    @classmethod
    def load_from_yaml(cls, yaml_file: Union[Path, str]) -> 'Configuration':
        """
        Loads configuration from a yaml file.

        **Returns:** `Configuration`

        Usage:

        ```yaml
        # conf.yaml
        scalpel:
          fetch_timeout: 4.0
          user_agent: Mozilla/5.0
          follow_robots_txt: true
        ```

        ```
        from scalpel import Configuration
        conf = Configuration.load_from_yaml('conf.yaml')
        conf.fetch_timeout  # equals to 4.0
        ```
        """
        cls._check_file(yaml_file, 'yaml')

        configuror = Config(mapping_files={'yaml': [f'{yaml_file}']})
        logger.debug('loading configuration from yaml file: %s', f'{yaml_file}')
        return cls(**cls._scalpel_attributes(configuror))

    @classmethod
    def load_from_toml(cls, toml_file: Union[Path, str]) -> 'Configuration':
        """
        Loads configuration from a toml file.

        **Returns:** `Configuration`

        Usage:

        ```toml
        # conf.toml
        [scalpel]
        user_agent = "Mozilla/5.0"
        fetch_timeout = 4.0
        follow_robots_txt = true
        ```

        ```
        from scalpel import Configuration
        conf = Configuration.load_from_toml('conf.toml')
        conf.fetch_timeout  # equals to 4.0
        ```
        """
        cls._check_file(toml_file, 'toml')

        configuror = Config(mapping_files={'toml': [f'{toml_file}']})
        logger.debug('loading configuration from toml file: %s', f'{toml_file}')
        return cls(**cls._scalpel_attributes(configuror))

    @classmethod
    def load_from_dotenv(cls, env_file: Union[Path, str]) -> 'Configuration':
        """
        Loads configuration from a .env file.

        **Returns:** `Configuration`

        Usage:

        ```bash
        # .env
        SCALPEL_USER_AGENT = Mozilla/5.0
        SCALPEL_FETCH_TIMEOUT = 4.0
        SCALPEL_FOLLOW_ROBOTS_TXT = yes
        ```

        ```
        from scalpel import Configuration
        conf = Configuration.load_from_dotenv('.env')
        conf.follow_robots_txt  # equals to True
        ```
        """
        cls._check_file(env_file, 'env')

        configuror = Config(mapping_files={'env': [f'{env_file}']})
        data = configuror.get_dict_from_namespace('SCALPEL_')
        data = {'scalpel': data}  # little trick to search attributes using _scalpel_attributes class method
        logger.debug('loading configuration from .env file: %s', f'{env_file}')
        return cls(**cls._scalpel_attributes(data))



import logging
import platform
from time import time
from typing import Callable, Optional, Any

import attr
import gevent
import httpx
from gevent.lock import RLock
from gevent.pool import Pool
from gevent.queue import JoinableQueue
from rfc3986 import uri_reference

from scalpel.core.spider import Spider
from .files import write_mp
from .response import StaticResponse
from .robots import RobotsAnalyzer
from .utils.io import open_file

logger = logging.getLogger('scalpel')


@attr.s(slots=True)
class StaticSpider(Spider):
    """
    A spider suitable to parse files or static HTML files.

    **Parameters:**

    * **urls:** Urls to parse. Allowed schemes are `http`, `https` and `file`. It can be a `list`, a `tuple` or a `set`.
    * **parse:** A callable used to parse url content. It takes two arguments: the current spider and a
    `StaticResponse` object.
    * **reachable_urls:** `set` of urls that are already fetched or read.
    * **unreachable_urls:** `set` of urls that were impossible to fetch or read.
    * **robot_excluded_urls:** `set` of urls that were excluded to fetch because of *robots.txt* file rules.
    * **followed_urls:** `set` of urls that were followed during the process of parsing url content. You will find these
    urls scattered in the first three sets.
    * **request_counter:** The number of urls already fetched or read.

    Usage:

    ```
    from scalpel.green import StaticSpider, StaticResponse

    def parse(spider: StaticSpider, response: StaticResponse) -> None:
        ...

    spider = StaticSpider(urls=['http://example.com'], parse=parse)
    spider.run()
    ```
    """
    # order is important here, http_client must come before robots_analyzer since the latter used the former
    _start_time: float = attr.ib(init=False, factory=time, repr=False)
    _http_client: httpx.Client = attr.ib(init=False, repr=False)
    _robots_analyser: RobotsAnalyzer = attr.ib(init=False, repr=False)
    _fetch: Callable = attr.ib(init=False, repr=False)
    _lock: RLock = attr.ib(factory=RLock, init=False, repr=False)
    _pool: Pool = attr.ib(factory=Pool, init=False, repr=False)
    _queue: JoinableQueue = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self):
        def _get_fetch(url: str) -> httpx.Response:
            return self._http_client.get(url)

        self._fetch = _get_fetch
        for middleware in self.config.response_middlewares:
            self._fetch = middleware(self._fetch)

    @_http_client.default
    def _get_http_client(self) -> httpx.Client:
        headers = {'User-Agent': self.config.user_agent}
        logger.debug('getting a default httpx client with user agent: %s', self.config.user_agent)
        return httpx.Client(headers=headers, timeout=self.config.fetch_timeout)

    @_robots_analyser.default
    def _get_robots_analyzer(self) -> RobotsAnalyzer:
        logger.debug('getting a default robots analyzer')
        return RobotsAnalyzer(
            http_client=self._http_client,
            robots_cache=self.config.robots_cache_folder,
            user_agent=self.config.user_agent
        )

    @_queue.default
    def _get_joinable_queue(self) -> JoinableQueue:
        logger.debug('getting a default joinable queue')
        return JoinableQueue(items=self.urls)

    def _get_static_response(
            self, url: str = '', text: str = '', httpx_response: httpx.Response = None
    ) -> StaticResponse:
        logger.debug(
            'returning StaticResponse object with url: %s, text: %s and httpx_response: %s', url, text, httpx_response
        )
        return StaticResponse(
            self.reachable_urls, self.followed_urls, self._queue, url=url, text=text, httpx_response=httpx_response
        )

    def _is_url_already_processed(self, url: str) -> bool:
        processed = False
        if url in [*self.reachable_urls, *self.unreachable_urls, *self.robots_excluded_urls]:
            logger.debug('url %s has already been processed', url)
            processed = True
        return processed

    def _is_url_excluded_for_spider(self, url: str) -> bool:
        excluded = False
        if self.config.follow_robots_txt:
            if not self._robots_analyser.can_fetch(url):
                logger.info('robots.txt rule has forbidden the processing of url %s or the url is not reachable', url)
                self.robots_excluded_urls.add(url)
                excluded = True
        return excluded

    # noinspection PyBroadException
    def _handle_url(self, url: str) -> None:
        if self._is_url_already_processed(url):
            return

        static_url = text = ''
        response: Optional[httpx.Response] = None
        ur = uri_reference(url)
        if ur.scheme == 'file':
            static_url = url
            logger.debug('url %s is a file url so we attempt to read its content')
            file_path = ur.path[1:] if platform.system() == 'Windows' else ur.path
            try:
                before = time()
                with open_file(file_path) as f:
                    text = f.read()
                fetch_time = time() - before
            except OSError:
                logger.exception('unable to open file %s', url)
                self.unreachable_urls.add(url)
                return
        else:
            if self._is_url_excluded_for_spider(url):
                return

            response: httpx.Response = self._fetch(url)
            if response.is_error:
                logger.info('fetching url %s returns an error with status code %s', url, response.status_code)
                self.unreachable_urls.add(url)
                return
            fetch_time = response.elapsed.total_seconds()

        # we update some variables for statistics
        self.request_counter += 1
        self.reachable_urls.add(url)
        self._total_fetch_time += fetch_time

        try:
            self.parse(self, self._get_static_response(static_url, text, response))
        except Exception:
            logger.exception('something unexpected happened while parsing the content at url %s', url)
            if not self._ignore_errors:
                raise
        logger.info('content at url %s has been processed', url)

    def save_item(self, item: Any) -> None:
        """Saves a scrapped item in the backup filename specified in `Configuration.backup_filename` attribute."""
        item_rejected = False
        original_item = item
        for processor in self.config.item_processors:
            item = processor(item)
            if item is None:
                item_rejected = True
                break
        if item_rejected:
            logger.debug('item %s was rejected', original_item)
            return

        logger.debug('writing item %s to file %s', item, self.config.backup_filename)
        with self._lock:
            write_mp(self.config.backup_filename, item, mode='a', encoder=self.config.msgpack_encoder)

    @staticmethod
    def _error_callback(task: gevent.Greenlet) -> None:
        if task.name == 'worker':
            logger.error('an unexpected error happened on the worker task', exc_info=task.exc_info)
        raise task.exception

    # NOTE: it is not possible to monkeypatch task_done method with pytest since JoinableQueue is a c-extension
    # this is why you will not see direct test of this method
    def _done_callback(self, *_) -> None:
        logger.debug('running greenlet done callback')
        self._queue.task_done()

    def _worker(self) -> None:
        # TODO: I had a weird LoopExit issue (while testing) when I tried to get the delay between requests
        #  and skipped some urls in the while loop, so to avoid it, I don't relay on robots.txt delay but on the one
        #  provided by config object.
        while True:
            task = self._pool.spawn(self._handle_url, self._queue.get())
            task.link_exception(self._error_callback)
            task.link(self._done_callback)
            gevent.sleep(self.config.request_delay)

    def _cleanup(self) -> None:
        """This method helps to cleanup resources. It should be override by SeleniumSpider."""
        self._http_client.close()

    def run(self) -> None:
        """Runs the spider."""
        worker_task = self._pool.spawn(self._worker)
        worker_task.name = 'worker'
        worker_task.link_exception(self._error_callback)
        self._queue.join()
        # at this point all urls were handled, so the only remaining task in the pool is the worker
        self._pool.killone(worker_task)
        self._cleanup()
        self._duration = time() - self._start_time



from .__version__ import __version__, version_info
from .core.config import Configuration, Browser
from .core.spider import State, SpiderStatistics
from .core.message_pack import datetime_encoder, datetime_decoder

__all__ = [
    '__version__', 'version_info',
    # config
    'Configuration', 'Browser',
    # message pack
    'datetime_encoder', 'datetime_decoder',
    # spider
    'State', 'SpiderStatistics'
]



import logging
from datetime import datetime
from typing import Union, Set, List, Tuple, Callable

import attr
from rfc3986 import iri_reference, validators, exceptions

from .config import Configuration

logger = logging.getLogger('scalpel')

URLS = Union[List[str], Tuple[str], Set[str]]


def url_validator(_, attribute: attr.Attribute, urls: URLS):
    if not isinstance(urls, (set, list, tuple)):
        message = f'{attribute.name} is not a set, list or tuple instance: {urls}'
        logger.exception(message)
        raise TypeError(message)

    if not all(isinstance(url, str) for url in urls):
        message = f'not all items in urls are a string: {urls}'
        logger.exception(message)
        raise TypeError(message)

    validator = validators.Validator()
    allowed_schemes = ['https', 'http', 'file']
    validator.allow_schemes(*allowed_schemes)
    validator.check_validity_of('scheme', 'host', 'path', 'query', 'fragment')

    for url in urls:
        uri = iri_reference(url).encode()
        try:
            validator.validate(uri)
        except exceptions.UnpermittedComponentError:
            message = f'{url} does not have a scheme in {allowed_schemes}'
            logger.exception(message)
            raise ValueError(message)
        except exceptions.InvalidComponentsError:
            # not sure this error can happen if we use uri_reference with a string, but let's be careful
            message = f'{url} is not a valid url'
            logger.exception(message)
            raise ValueError(message)

        if uri.scheme in ('http', 'https') and uri.host is None:
            message = f'url {url} must provide a host part'
            logger.exception(message)
            raise ValueError(message)

        if uri.scheme == 'file' and uri.path is None:
            message = f'url {url} must provide a path to a local file'
            logger.exception(message)
            raise ValueError(message)


@attr.s
class State:
    """An empty class used to store arbitrary data."""
    pass


@attr.s(frozen=True)
class SpiderStatistics:
    """
    Provides some statistics about a ran spider.

    **Parameters:**

    * **reachable_urls:** `set` of urls that were fetched (or read in case of file urls) and parsed.
    * **unreachable_urls:** `set` that were impossible to fetch (or read in case of file urls).
    * **robot_excluded_urls:** `set` of urls that were excluded to fetch because of *robots.txt* file rules.
    * **followed_urls:** `set` of urls that were followed during the process of parsing url content. You will find these
    urls scattered in the first three sets.
    * **request_counter:** The number of urls fetched or read (in case of file urls).
    * **average_fetch_time:** The average time to fetch an url (or read a file in case of file urls).
    * **total_time:** The total execution time of the spider.
    """
    reachable_urls: Set[str] = attr.ib()
    unreachable_urls: Set[str] = attr.ib()
    robot_excluded_urls: Set[str] = attr.ib()
    followed_urls: Set[str] = attr.ib()
    request_counter: int = attr.ib()
    average_fetch_time: float = attr.ib()
    total_time: float = attr.ib()


@attr.s(slots=True)
class Spider:
    urls: URLS = attr.ib(validator=url_validator)
    parse: Callable = attr.ib(validator=attr.validators.is_callable())
    _name: str = attr.ib(validator=attr.validators.instance_of(str))
    _config: Configuration = attr.ib(
        factory=Configuration, validator=attr.validators.instance_of(Configuration), repr=False
    )
    _ignore_errors: bool = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    reachable_urls: Set[str] = attr.ib(factory=set, init=False, repr=False)
    unreachable_urls: Set[str] = attr.ib(factory=set, init=False, repr=False)
    followed_urls: Set[str] = attr.ib(factory=set, init=False, repr=False)
    robots_excluded_urls: Set[str] = attr.ib(factory=set, init=False, repr=False)
    request_counter: int = attr.ib(default=0, init=False, repr=False)
    _duration: float = attr.ib(init=False, default=0.0, repr=False)
    _total_fetch_time: float = attr.ib(init=False, default=0.0)
    _state: State = attr.ib(factory=State, init=False, repr=False)

    @_name.default
    def _get_name(self) -> str:
        now = datetime.now()
        name = f'spider-{now.strftime("%Y-%m-%d@%H:%M:%S.%f")}'
        logger.debug('getting a default spider name: %s', name)
        return name

    @property
    def name(self) -> str:
        """Returns the name given to the spider."""
        logger.debug('returning name property: %s', self._name)
        return self._name

    @property
    def config(self) -> Configuration:
        """Returns the `Configuration` related to the spider."""
        logger.debug('returning config property: %s', self._config)
        return self._config

    @property
    def state(self) -> State:
        """Returns the `State` related to the spider. You can add custom information on this object."""
        logger.debug('returning state property: %s', self._state)
        return self._state

    def statistics(self) -> SpiderStatistics:
        """
        Provides some statistics related to the ran spider.

        **Returns:** `SpiderStatistics`
        """
        return SpiderStatistics(
            reachable_urls=self.reachable_urls,
            unreachable_urls=self.unreachable_urls,
            robot_excluded_urls=self.robots_excluded_urls,
            followed_urls=self.followed_urls,
            request_counter=self.request_counter,
            average_fetch_time=self._total_fetch_time / self.request_counter if self.request_counter else 0.0,
            total_time=self._duration
        )



import logging

import attr
from rfc3986 import uri_reference

from scalpel.core.selenium import SeleniumDriverMixin
from .mixins import SeleniumGetMixin
from .response import SeleniumResponse
from .static_spider import StaticSpider

logger = logging.getLogger('scalpel')


# usually mixins are put are the beginning but in the case of SeleniumDriverMixin
# it depends on _config property defined in StaticSpider, so it has to be put at the end
@attr.s
class SeleniumSpider(SeleniumGetMixin, StaticSpider, SeleniumDriverMixin):
    """
    A spider suitable to parse dynamic websites i.e where Javascript is heavily used. You will sometimes encounter the
    term Single-Page Application (SPA) for this type of website. It relies on `selenium` package and a browser.

    **Parameters:**

    * **urls:** Urls to parse. Allowed schemes are `http`, `https` and `file`. It can be a `list`, a `tuple` or a `set`.
    * **parse:** An async callable used to parse url content. It takes two arguments: the current spider and a
    `StaticResponse` object.
    * **reachable_urls:** `set` of urls that are already fetched or read.
    * **unreachable_urls:** `set` of urls that were impossible to fetch or read.
    * **robot_excluded_urls:** `set` of urls that were excluded to fetch because of *robots.txt* file rules.
    * **followed_urls:** `set` of urls that were followed during the process of parsing url content. You will find these
    urls scattered in the first three sets.
    * **request_counter:** The number of urls already fetched or read.

    Usage:

    ```
    from scalpel.any_io import SeleniumSpider, SeleniumResponse

    async def parse(spider: SeleniumSpider, response: SeleniumResponse) -> None:
        ...

    spider = SeleniumSpider(urls=['https://example.com'], parse=parse)
    await spider.run()
    ```
    """

    def __attrs_post_init__(self):
        # this erases the content of StaticSpider.__attrs_post_init__ which is not useful here
        pass

    def _get_selenium_response(self, handle: str) -> SeleniumResponse:
        return SeleniumResponse(
            self.reachable_urls, self.followed_urls, self._queue, driver=self._driver, handle=handle
        )

    async def _cleanup(self) -> None:
        self._driver.quit()
        await self._http_client.aclose()
        await self._queue.close()

    # noinspection PyBroadException
    async def _handle_url(self, url: str) -> None:
        if self._is_url_already_processed(url):
            return

        ur = uri_reference(url)
        error_message = ''
        if ur.scheme == 'file':
            error_message = f'unable to open file {url}'

        unreachable, fetch_time = self._get_resource(url, error_message)
        if unreachable:
            self.unreachable_urls.add(url)
            self._queue.task_done()
            return
        # we update some stats
        self.request_counter += 1
        self._total_fetch_time += fetch_time
        self.reachable_urls.add(url)

        handle = self._driver.current_window_handle
        try:
            await self.parse(self, self._get_selenium_response(handle))
        except Exception:
            logger.exception('something unexpected happened while parsing the content at url %s', url)
            if not self._ignore_errors:
                self._queue.task_done()
                raise
        self._queue.task_done()
        logger.info('content at url %s has been processed', url)



"""Utility async functions to read and write mp files."""
import logging
from pathlib import Path
from typing import Union, AsyncIterator, Any, Callable

import anyio
import msgpack

logger = logging.getLogger('scalpel')


async def read_mp(filename: Union[str, Path], decoder: Callable = None) -> AsyncIterator[Any]:
    """
    Reads a `msgpack` file generated by the spider when calling the `save_item` method.

    **Parameters:**

    * **filename:** The name of the file to read. It can be a string or a `pathlib.Path`.
    * **decoder:** An optional function used to decode data types not handled by default by `msgpack`.

    Usage:

    ```python
    from scalpel import datetime_decoder
    from scalpel.any_io import read_mp

    async for item in read_mp('file.mp', datetime_decoder):
        print(item)
    ```
    """
    if decoder is not None and not callable(decoder):
        message = f'{decoder} is not callable'
        logger.exception(message)
        raise TypeError(message)

    async with await anyio.open_file(filename, 'rb') as f:
        content = await f.read()
        unpacker = msgpack.Unpacker(None, max_buffer_size=len(content), object_hook=decoder)
        unpacker.feed(content)
        logger.debug('reading data from file %s', filename)
        for data in unpacker:
            yield data


async def write_mp(filename: Union[str, Path], data: Any, mode: str = 'w', encoder: Callable = None) -> int:
    """
    Writes a `msgpack` file.

    **Parameters:**

    * **filename:** The name of the file where data will be written. It can be a string or a `pathlib.Path`.
    * **data:** Arbitrary data to serialize. Note that if you want to serialize data types not supported by the `json`
    module, you will need to provide a custom encoder function.
    * **mode:** The mode in which the file is opened. Valid values are "a" (append) and "w" (write). Defaults to "a".
    * **encoder:** An optional function used to encode data types not handled by default by `msgpack`.

    **Returns:** The number of written bytes.

    Usage:

    ```
    from datetime import datetime
    from scalpel import datetime_encoder
    from scalpel.any_io import write_mp

    data = {'fruit': 'apple', 'date': datetime.utcnow()}
    length = await write_mp('file.mp', data, 'w', datetime_encoder)
    print(length)  # 65
    ```
    """
    if mode not in ['a', 'w']:
        message = f'The only modes expected are "a" and "w" but you provided {mode}'
        logger.exception(message)
        raise TypeError(message)

    if encoder is not None and not callable(encoder):
        message = f'{encoder} is not callable'
        logger.exception(message)
        raise TypeError(message)

    async with await anyio.open_file(filename, f'{mode}b') as f:
        data_length = await f.write(msgpack.packb(data, default=encoder))
        logger.debug('writing %s bytes in file %s', data_length, filename)
        return data_length



import logging
from typing import Set

import attr
from gevent.queue import JoinableQueue

from scalpel.core.response import BaseStaticResponse, BaseSeleniumResponse

logger = logging.getLogger('scalpel')


class FollowMixin:
    def follow(self, url: str) -> None:
        """
        Follows given url if it hasn't be fetched yet.

        **Parameters:**

        * **url:** The url to follow.
        """
        if url in self._followed_urls or url in self._reachable_urls:
            logger.debug('url %s has already been processed, nothing to do here', url)
            return

        logger.debug('adding url %s to spider followed_urls attribute and put in the queue to be processed', url)
        url = self._get_absolute_url(url)
        self._followed_urls.add(url)
        self._queue.put_nowait(url)


@attr.s
class CommonAttributes:
    _reachable_urls: Set[str] = attr.ib(validator=attr.validators.instance_of(set))
    _followed_urls: Set[str] = attr.ib(validator=attr.validators.instance_of(set))
    _queue: JoinableQueue = attr.ib(validator=attr.validators.instance_of(JoinableQueue))


@attr.s(slots=True)
class StaticResponse(CommonAttributes, FollowMixin, BaseStaticResponse):
    """
    A response class used in combination with a `StaticSpider` object in the `parse` callable of a spider.

    **N.B:** You probably don't need to instantiate this class directly unless for some kind of testing. It is mainly
    exposed for annotation purpose.

    **Parameters:**

    * **reachable_urls:** A `set` of urls already fetched.
    * **followed_urls:** A `set` of urls already followed by other `StaticResponse` objects.
    * **queue:** The `gevent.queue.JoinableQueue` used by the spider to handle incoming urls.
    * **url:** An optional keyword parameter representing the current url where content was fetched.
    * **text:** An optional keyword parameter representing the content of the resource fetched. Note that if you set
    the `url` parameter, you **must** set this one.
    * **httpx_response:** An optional keyword parameter representing an `httpx.Response` of the resource fetched.
    For HTTP urls, this is the one used in favour of `url` and `text` parameters.

    Usage:

    ```
    from scalpel.green import StaticResponse

    response = StaticResponse(..., url='http://foo.com', text='<p>Hello world!</p>')
    print(response.css('p::text').get())  # 'Hello world!'
    print(response.xpath('//p/text()').get())  # 'Hello world!'
    ```
    """
    pass


@attr.s(slots=True)
class SeleniumResponse(CommonAttributes, FollowMixin, BaseSeleniumResponse):
    """
    A response class used in combination with a `SeleniumSpider` object in the `parse` callable of a spider.

    **N.B:** You probably don't need to instantiate this class directly unless for some kind of testing. It is mainly
    exposed for annotation purpose.

    **Parameters:**

    * **reachable_urls:** A `set` of urls already fetched.
    * **followed_urls:** A `set` of urls already followed by other `StaticResponse` objects.
    * **queue:** The `gevent.queue.JoinableQueue` used by the spider to handle incoming urls.
    * **driver:** The `selenium.WebDriver` object that will be use to control the running browser.
    * **handle:** A string that identifies the current window handled by `selenium`.

    Usage:

    ```
    from scalpel.green import SeleniumResponse

    response = SeleniumResponse(...)
    # We assume we have a page source like '<p>Hello world!</p>'
    print(response.driver.find_element_by_xpath('//p').text)  # Hello world!
    ```
    """
    pass



import logging
from typing import Tuple

import anyio
from selenium.common.exceptions import WebDriverException

logger = logging.getLogger('scalpel')


class SeleniumGetMixin:

    def _get_resource(self, url: str, error_message: str = '') -> Tuple[bool, float]:
        fetch_time = 0
        unreachable = False
        if not error_message:
            error_message = f'unable to get resource at {url}'
        try:
            before = anyio.current_time()
            self._driver.get(url)
            fetch_time = anyio.current_time() - before
        except WebDriverException:
            logger.exception(error_message)
            unreachable = True

        return unreachable, fetch_time



from .files import read_mp, write_mp
from .response import StaticResponse, SeleniumResponse
from .selenium_spider import SeleniumSpider
from .static_spider import StaticSpider
from .queue import Queue

__all__ = [
    # files
    'read_mp', 'write_mp',
    # spider
    'StaticSpider', 'SeleniumSpider',
    # response
    'StaticResponse', 'SeleniumResponse',
    # queue
    'Queue'
]



import logging
from typing import Dict, Union
from urllib.robotparser import RobotFileParser

logger = logging.getLogger(__name__)


class RobotsMixin:

    # noinspection PyTypeChecker
    @staticmethod
    def _get_request_delay(
            host: str,
            url: str,
            robots_parser: RobotFileParser,
            delay_mapping: Dict[str, Union[int, float]],
            default_delay: Union[int, float]
    ) -> Union[int, float]:
        pass

        crawl_delay = robots_parser.crawl_delay('*')
        if crawl_delay is not None:
            delay_mapping[host] = crawl_delay
            logger.debug('returning crawl delay value "%s" from robots.txt for url %s', crawl_delay, url)
            return crawl_delay

        request_rate = robots_parser.request_rate('*')
        if request_rate is not None:
            request_delay = request_rate.seconds / request_rate.requests
            delay_mapping[host] = request_delay
            logger.debug(
                'computing value "%s" from request delay info (%s/%s) from robots.txt for url %s',
                request_delay, request_rate.requests, request_rate.seconds, url
            )
            return request_delay

        delay_mapping[host] = default_delay
        logger.debug('returning default delay value "%s" for url %s', default_delay, url)
        return default_delay



"""Implementation of an anyio queue."""
import logging
from typing import Iterable, Any, Union

import anyio
import attr
from anyio.streams.memory import MemoryObjectSendStream, MemoryObjectReceiveStream

logger = logging.getLogger('scalpel')


@attr.s
class Queue:
    """
    An implementation of an anyio queue with the capacity to tell that a task is done.

    **Parameters:**

    * **size:** The size of the queue, defaults to 1 (this is the minimum value).
    * **items:** A list of values used to initialize the queue.

    **N.B:** If you set `items`, be sure that the size of the queue is greater than `items` length, otherwise you will
    have a `anyio.WouldBlock` exception.

    ```python
    import math
    import random

    import anyio
    from scalpel.any_io import Queue

    async def worker(name: str, queue: Queue) -> None:
        while True:
            # Get a "work item" out of the queue.
            time_to_sleep = await queue.get()

            await trio.sleep(time_to_sleep)
            # Notify the queue that the "work item" has been processed.
            queue.task_done()
            print(f'{name} has slept for {time_to_sleep:.2f} seconds')


    async def main():
        async with Queue(math.inf) as queue:
            # Generate random timings and put them into the queue.
            total_sleep_time = 0
            for _ in range(20):
                sleep_for = random.uniform(0.05, 1.0)
                total_sleep_time += sleep_for
                queue.put_nowait(sleep_for)

            # Create three worker tasks to process the queue concurrently.
            async with anyio.create_task_group() as tg:
                for i in range(3):
                    tg.spawn(worker, f'worker-{i}', queue)

                # Wait until the queue is fully processed.
                before = anyio.current_time()
                await queue.join()
                total_slept_for = trio.current_time() - before
                print('====')
                print(f'3 workers slept in parallel for {total_slept_for:.2f} seconds')
                print(f'total expected sleep time: {total_sleep_time:.2f} seconds')

                # We can end the nursery which will in turn end the tasks.
                tg.cancel_scope.cancel()


    anyio.run(main)
    ```
    """
    _size: Union[int, float] = attr.ib(default=1, validator=attr.validators.instance_of((int, float)))
    _items: Iterable[Any] = attr.ib(factory=list, validator=attr.validators.instance_of((list, set, tuple)))
    _send_channel: MemoryObjectSendStream = attr.ib(init=False)
    _receive_channel: MemoryObjectReceiveStream = attr.ib(init=False)
    _tasks_in_progress: int = attr.ib(default=0, init=False)
    _finished: anyio.Event = attr.ib(factory=anyio.Event, init=False)

    def __attrs_post_init__(self):
        logger.debug('initializing queue send and receive channels with a size of %s', self._size)
        self._send_channel, self._receive_channel = anyio.create_memory_object_stream(self._size)
        if self._items:
            logger.debug('trying to add items coming from list %s', self._items)
            for item in self._items:
                self._tasks_in_progress += 1
                self._send_channel.send_nowait(item)

    @_size.validator
    def _validate_size(self, _, value: int) -> None:
        if value < 1:
            message = f'size must not be less than 1 but you provide: {value}'
            logger.error(message)
            raise ValueError(message)

    @property
    def length(self) -> int:
        """Tne number of items currently present in the queue."""
        value = self._send_channel.statistics().current_buffer_used
        logger.debug('returning queue length: %s', value)
        return value

    @property
    def maxsize(self) -> Union[int, float]:
        """The number of items allowed in the queue."""
        logger.debug('returning queue max size: %s', self._size)
        return self._size

    async def put(self, item: Any) -> None:
        logger.debug('adding item %s to queue', item)
        await self._send_channel.send(item)
        self._tasks_in_progress += 1
        logger.debug('number of tasks in progress is now: %s', self._tasks_in_progress)
        if self._finished.is_set():
            self._finished = anyio.Event()

    def put_nowait(self, item: Any) -> None:
        logger.debug('trying to add %s to queue', item)
        self._send_channel.send_nowait(item)
        self._tasks_in_progress += 1
        logger.debug('number of tasks in progress is now: %s', self._tasks_in_progress)
        if self._finished.is_set():
            self._finished = anyio.Event()

    async def get(self) -> Any:
        item = await self._receive_channel.receive()
        logger.debug('returning item %s from queue', item)
        return item

    def get_nowait(self) -> Any:
        logger.debug('trying to get item from queue')
        return self._receive_channel.receive_nowait()

    def task_done(self) -> None:
        if self._tasks_in_progress <= 0:
            raise ValueError('task_done method was calling too many times without adding items in queue')
        self._tasks_in_progress -= 1
        logger.debug('decrementing tasks in progress, now the new value is: %s', self._tasks_in_progress)
        if self._tasks_in_progress == 0:
            self._finished.set()

    async def close(self) -> None:
        logger.debug('closing send and receive channels')
        await self._send_channel.aclose()
        await self._receive_channel.aclose()

    async def join(self) -> None:
        logger.debug('waiting self._tasks_in_progress to be equal to 0')
        await self._finished.wait()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()



"""Helper datetime encoder and decoder for msgpack"""
from datetime import datetime
from typing import Any


def datetime_encoder(data: Any) -> Any:
    """
    A datetime encoder for `msgpack`

    Usage:

    ```
    from datetime import datetime
    from scalpel import datetime_encoder
    import msgpack

    data = {'fruit': 'apple', 'date': datetime.utcnow()}
    packed_data = msgpack.packb(data, default=datetime_encoder)
    ```
    """
    if isinstance(data, datetime):
        return {'__datetime__': True, 'as_str': data.strftime('%Y%m%dT%H:%M:%S.%f')}
    return data  # pragma: no cover


def datetime_decoder(data: Any) -> Any:
    """
    A datetime decoder for `msgpack`.

    Usage:

    ```
    from datetime import datetime
    from scalpel import datetime_encoder, datetime_decoder
    import msgpack

    data = {'fruit': 'apple', 'date': datetime.utcnow()}
    packed_data = msgpack.packb(data, default=datetime_encoder)
    assert msgpack.unpackb(packed_data, object_hook=datetime_decoder) == data
    ```
    """
    if '__datetime__' in data:
        data = datetime.strptime(data['as_str'], '%Y%m%dT%H:%M:%S.%f')
    return data



import logging
import math
import platform
from asyncio import iscoroutinefunction
from pathlib import Path
from typing import Callable, Optional, Any, Union

import anyio
import attr
import httpx
from anyio.abc import TaskGroup
from rfc3986 import uri_reference

from scalpel.core.spider import Spider
from .files import write_mp
from .response import StaticResponse
from .robots import RobotsAnalyzer
from .queue import Queue

logger = logging.getLogger('scalpel')


@attr.s(slots=True)
class StaticSpider(Spider):
    """
    A spider suitable to parse files or static HTML files.

    **Parameters:**

    * **urls:** Urls to parse. Allowed schemes are `http`, `https` and `file`. It can be a `list`, a `tuple` or a `set`.
    * **parse:** An async callable used to parse url content. It takes two arguments: the current spider and a
    `StaticResponse` object.
    * **reachable_urls:** `set` of urls that are already fetched or read.
    * **unreachable_urls:** `set` of urls that were impossible to fetch or read.
    * **robot_excluded_urls:** `set` of urls that were excluded to fetch because of *robots.txt* file rules.
    * **followed_urls:** `set` of urls that were followed during the process of parsing url content. You will find these
    urls scattered in the first three sets.
    * **request_counter:** The number of urls already fetched or read.

    Usage:

    ```
    from scalpel.any_io import StaticSpider, StaticResponse

    async def parse(spider: StaticSpider, response: StaticResponse) -> None:
        ...

    spider = StaticSpider(urls=['https://example.com'], parse=parse)
    await spider.run()
    ```
    """
    # order is important here, http_client must come before robots_analyzer since the latter used the former
    _start_time: float = attr.ib(init=False, repr=False, factory=anyio.current_time)
    _http_client: httpx.AsyncClient = attr.ib(init=False, repr=False)
    _robots_analyser: RobotsAnalyzer = attr.ib(init=False, repr=False)
    _fetch: Callable = attr.ib(init=False, repr=False)
    _lock: anyio.Lock = attr.ib(init=False, repr=False, factory=anyio.Lock)
    _queue: Queue = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self):

        async def _get_fetch(url: str) -> httpx.Response:
            return await self._http_client.get(url)

        self._fetch = _get_fetch
        for middleware in self.config.response_middlewares:
            self._fetch = middleware(self._fetch)

    @_http_client.default
    def _get_http_client(self) -> httpx.AsyncClient:
        headers = {'User-Agent': self.config.user_agent}
        logger.debug('getting a default httpx client with user agent: %s', self.config.user_agent)
        return httpx.AsyncClient(headers=headers, timeout=self.config.fetch_timeout)

    @_robots_analyser.default
    def _get_robots_analyzer(self) -> RobotsAnalyzer:
        logger.debug('getting a default robots analyzer')
        return RobotsAnalyzer(
            http_client=self._http_client,
            robots_cache=Path(self.config.robots_cache_folder),
            user_agent=self.config.user_agent
        )

    @_queue.default
    def _get_queue(self) -> Queue:
        logger.debug('getting a default queue')
        return Queue(size=math.inf, items=self.urls)

    def _get_static_response(
            self, url: str = '', text: str = '', httpx_response: httpx.Response = None
    ) -> StaticResponse:
        logger.debug(
            'returning StaticResponse object with url: %s, text: %s and httpx_response: %s', url, text, httpx_response
        )
        return StaticResponse(
            self.reachable_urls, self.followed_urls, self._queue, url=url, text=text,
            httpx_response=httpx_response
        )

    def _is_url_already_processed(self, url: str) -> False:
        processed = False
        if url in [*self.reachable_urls, *self.unreachable_urls, *self.robots_excluded_urls]:
            logger.debug('url %s has already been processed', url)
            self._queue.task_done()
            processed = True
        return processed

    # noinspection PyBroadException
    async def _handle_url(self, url: str) -> None:
        if self._is_url_already_processed(url):
            return

        static_url = text = ''
        response: Optional[httpx.Response] = None
        ur = uri_reference(url)
        if ur.scheme == 'file':
            static_url = url
            logger.debug('url %s is a file url so we attempt to read its content')
            file_path = ur.path[1:] if platform.system() == 'Windows' else ur.path
            try:
                before = anyio.current_time()
                async with await anyio.open_file(file_path) as f:
                    text = await f.read()
                fetch_time = anyio.current_time() - before
            except OSError:
                logger.exception('unable to open file %s', url)
                self.unreachable_urls.add(url)
                return
        else:
            response: httpx.Response = await self._fetch(url)
            if response.is_error:
                logger.info('fetching url %s returns an error with status code %s', url, response.status_code)
                self.unreachable_urls.add(url)
                return
            fetch_time = response.elapsed.total_seconds()

        # we update some variables for statistics
        self.request_counter += 1
        self.reachable_urls.add(url)
        self._total_fetch_time += fetch_time

        try:
            await self.parse(self, self._get_static_response(static_url, text, response))
        except Exception:
            logger.exception('something unexpected happened while parsing the content at url %s', url)
            if not self._ignore_errors:
                self._queue.task_done()
                raise
        self._queue.task_done()
        logger.info('content at url %s has been processed', url)

    async def save_item(self, item: Any) -> None:
        item_rejected = False
        original_item = item
        for processor in self.config.item_processors:
            if iscoroutinefunction(processor):
                item = await processor(item)
            else:
                item = processor(item)
            if item is None:
                item_rejected = True
                break
        if item_rejected:
            logger.debug('item %s was rejected', original_item)
            return

        logger.debug('writing item %s to file %s', item, self.config.backup_filename)
        async with self._lock:
            await write_mp(self.config.backup_filename, item, mode='a', encoder=self.config.msgpack_encoder)

    async def _get_request_delay(self, url: str) -> Union[int, float]:
        if self.config.follow_robots_txt:
            return await self._robots_analyser.get_request_delay(url, self.config.request_delay)
        await anyio.sleep(0)  # checkpoint to ensure an async function in every case
        return self.config.request_delay

    async def worker(self, task_group: TaskGroup) -> None:
        while True:
            url = await self._queue.get()
            request_delay = await self._get_request_delay(url)
            if request_delay == -1:  # url is not accessible
                self.robots_excluded_urls.add(url)
                self._queue.task_done()
                continue

            task_group.start_soon(self._handle_url, url)
            await anyio.sleep(request_delay)

    async def _cleanup(self) -> None:
        await self._http_client.aclose()
        await self._queue.close()

    # noinspection PyAsyncCall
    async def run(self) -> None:
        """Runs the spider."""
        async with anyio.create_task_group() as tg:
            tg.start_soon(self.worker, tg)
            await self._queue.join()
            # at this point, all the urls were handled, so the only remaining task is the worker
            tg.cancel_scope.cancel()

        await self._cleanup()
        self._duration = anyio.current_time() - self._start_time



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scalpel',
 'scalpel.any_io',
 'scalpel.core',
 'scalpel.green',
 'scalpel.green.utils']

package_data = \
{'': ['*']}

install_requires = \
['anyio>=3.0.0,<4.0.0',
 'attrs>=19.3.0,<20.0.0',
 'configuror>=0.2.0,<0.3.0',
 'fake-useragent>=0.1.11,<0.2.0',
 'httpx>=0.17.0,<0.18.0',
 'msgpack>=1.0.0,<2.0.0',
 'parsel>=1.5.2,<2.0.0',
 'rfc3986[idna]>=1.4.0,<2.0.0',
 'selenium>=3.141.0,<4.0.0']

extras_require = \
{'full': ['gevent>=20.9.0,<21.0.0', 'trio>=0.17.0,<0.18.0'],
 'gevent': ['gevent>=20.9.0,<21.0.0'],
 'trio': ['trio>=0.17.0,<0.18.0']}

setup_kwargs = {
    'name': 'pyscalpel',
    'version': '0.1.2',
    'description': 'Your easy-to-use, fast and powerful web scraping library',
    'long_description': '# Pyscalpel\n\n[![Pypi version](https://img.shields.io/pypi/v/pyscalpel.svg)](https://pypi.org/project/pyscalpel/)\n![](https://github.com/lewoudar/scalpel/workflows/CI/badge.svg)\n[![Coverage Status](https://codecov.io/gh/lewoudar/scalpel/branch/master/graphs/badge.svg?branch=master)](https://codecov.io/gh/lewoudar/scalpel)\n[![Documentation Status](https://readthedocs.org/projects/scalpel/badge/?version=latest)](https://scalpel.readthedocs.io/en/latest/?badge=latest)\n[![License Apache 2](https://img.shields.io/hexpm/l/plug.svg)](http://www.apache.org/licenses/LICENSE-2.0)\n\nYour easy-to-use, fast and powerful web scraping library.\n\n## Why?\n\nI already knew [scrapy](https://docs.scrapy.org/en/latest/) which is the reference in python for web scraping. But\ntwo things bothered me.\n- I feel like scrapy cannot integrate into an existing project, you need to treat your web scraping stuff like a project\non its own.\n- Usage of [Twisted](https://twistedmatrix.com/trac/) who is a veteran in asynchronous programming, but I think\n that there are better asynchronous frameworks today. Note that this second point is not true anymore as I\'m writing\n the document since scrapy adds support for [asyncio](https://docs.scrapy.org/en/latest/topics/asyncio.html)\n \n After having made this observation I decided to create pyscalpel. And let\'s be honest, I also want to have my own web\n scraping library, and it is fun to write one ;)\n \n\n## Installation\n \n```bash\npip install pyscalpel  # # to only use the asyncio backend\npip install pyscalpel[gevent] # to install the gevent backend\npip install pyscalpel[trio] # to installl the trio backend\npip install pyscalpel[full] # to install all the backends\n```\n\nIf you know about [poetry](https://python-poetry.org/) you can use it instead of pip.\n\n```bash\npoetry add pyscalpel  # to only use the asyncio backend\npoetry add pyscalpel[gevent] # to install the gevent backend\npoetry add pyscalpel[trio] # to install the trio backend\npoetry add pyscalpel[full] # to install all the backends\n```\n\npyscalpel works starting from **python 3.6**, it relies on robust packages:\n- [configuror](https://configuror.readthedocs.io/en/latest/): A configuration toolkit. \n- [httpx](https://www.python-httpx.org/): A modern http client.\n- [selenium](https://pypi.org/project/selenium/): A library for controlling a browser.\n- [gevent](http://www.gevent.org/): An asynchronous framework using the synchronous way. (optional)\n- [trio](https://trio.readthedocs.io/en/stable/): A modern asynchronous framework using `async/await` syntax. (optional)\n- [anyio](https://anyio.readthedocs.io/): An asynchronous networking and concurrency library that works on top of\neither asyncio or trio.\n- [parsel](https://parsel.readthedocs.io/): A library elements in HTML/XML documents.\n- [attrs](https://www.attrs.org/en/stable/): A library helping to write classes without pain.\n- [fake-useragent](https://pypi.org/project/fake-useragent/): A simple library to fake a user agent.\n- [rfc3986](https://rfc3986.readthedocs.io/en/latest/): A library for url parsing and validation.\n- [msgpack](https://pypi.org/project/msgpack/): A library allowing for fast serialization/deserialization of data\nstructures.\n\n## Documentation\n\nThe documentation is available at https://scalpel.readthedocs.io/en/latest/.\n\n\n## Usage\n\nTo give you an overview of what can be done, this is a simple example of quote scraping. Don\'t hesitate to look at the\nexamples folder for more snippets to look at.\n\nwith gevent\n\n```python\nfrom pathlib import Path\n\nfrom scalpel import Configuration\nfrom scalpel.green import StaticSpider, StaticResponse, read_mp\n\ndef parse(spider: StaticSpider, response: StaticResponse) -> None:\n    for quote in response.xpath(\'//div[@class="quote"]\'):\n        data = {\n            \'message\': quote.xpath(\'./span[@class="text"]/text()\').get(),\n            \'author\': quote.xpath(\'./span/small/text()\').get(),\n            \'tags\': quote.xpath(\'./div/a/text()\').getall()\n        }\n        spider.save_item(data)\n\n    next_link = response.xpath(\'//nav/ul/li[@class="next"]/a\').xpath(\'@href\').get()\n    if next_link is not None:\n        response.follow(next_link)\n\nif __name__ == \'__main__\':\n    backup = Path(__file__).parent / \'backup.mp\'\n    config = Configuration(backup_filename=f\'{backup}\')\n    spider = StaticSpider(urls=[\'http://quotes.toscrape.com\'], parse=parse, config=config)\n    spider.run()\n    print(spider.statistics())\n    # you can do whatever you want with the results\n    for quote_data in read_mp(filename=backup, decoder=spider.config.msgpack_decoder):\n        print(quote_data)\n```\n\nwith anyio\n\n```python\nfrom pathlib import Path\n\nimport anyio\nfrom scalpel import Configuration\nfrom scalpel.any_io import StaticResponse, StaticSpider, read_mp\n\n\nasync def parse(spider: StaticSpider, response: StaticResponse) -> None:\n    for quote in response.xpath(\'//div[@class="quote"]\'):\n        data = {\n            \'message\': quote.xpath(\'./span[@class="text"]/text()\').get(),\n            \'author\': quote.xpath(\'./span/small/text()\').get(),\n            \'tags\': quote.xpath(\'./div/a/text()\').getall()\n        }\n        await spider.save_item(data)\n\n    next_link = response.xpath(\'//nav/ul/li[@class="next"]/a\').xpath(\'@href\').get()\n    if next_link is not None:\n        await response.follow(next_link)\n\nasync def main():\n    backup = Path(__file__).parent / \'backup.mp\'\n    config = Configuration(backup_filename=f\'{backup}\')\n    spider = StaticSpider(urls=[\'http://quotes.toscrape.com\'], parse=parse, config=config)\n    await spider.run()\n    print(spider.statistics())\n    # you can do whatever you want with the results\n    async for item in read_mp(backup, decoder=spider.config.msgpack_decoder):\n        print(item)\n\nif __name__ == \'__main__\':\n    # by default, this will run the asyncio backend, if you want the trio backend, you must first install the trio \n    # package and replace the following line with: anyio.run(main, backend=\'trio\').\n    anyio.run(main)\n```\n\n## Known limitations\n\npyscalpel aims to handle SPA (single page application) through the use of selenium. However, due to the synchronous nature\nof selenium, it is hard to leverage anyio and gevent asynchronous feature. You will notice that the *selenium spider* is\nslower than the *static spider*. For more information look at the documentation.\n\n## Warning\n\npyscalpel is a young project, so it is expected to have breaking changes in the api without respecting the \n[semver](https://semver.org/) principle. It is recommended to pin the version you are using for now.',
    'author': 'lewoudar',
    'author_email': 'lewoudar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://scalpel.readthedocs.io/en/stable',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from gevent import monkey

monkey.patch_all()
from .files import read_mp, write_mp  # noqa: E402
from .utils.io import wrap_file, open_file, AsyncFile  # noqa: E402
from .response import StaticResponse, SeleniumResponse  # noqa: E402
from .static_spider import StaticSpider  # noqa: E402
from .selenium_spider import SeleniumSpider  # noqa: E402

__all__ = [
    # files
    'read_mp', 'write_mp',
    # io
    'wrap_file', 'open_file', 'AsyncFile',
    # spider
    'StaticSpider', 'SeleniumSpider',
    # response
    'StaticResponse', 'SeleniumResponse'
]
